﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;
using QuadTreeLib;

namespace QuadTreeExample
{
    public partial class QuadTreeExample : Form
    {
        QuadTree<Item> _quadTree;
        List<Item> _items = new List<Item>();
        
        private int _index = 0;
        bool _dragging = false;
        QuadTreeLib.RectangleF _selectionRect;
        Point _startPoint;

        public QuadTreeExample()
        {
            InitializeComponent();
            Init();
        }

        private void Init()
        {
            _index = 0;
            _items.Clear();
            _quadTree = new QuadTree<Item>(this.pictureBox.ClientRectangle.X, this.pictureBox.ClientRectangle.Y, this.pictureBox.ClientRectangle.Width, this.pictureBox.ClientRectangle.Height,
                6, 1);
            _selectionRect = QuadTreeLib.RectangleF.Empty;
        }

        private Rectangle RoundReactangle(QuadTreeLib.RectangleF rect)
        {
            System.Drawing.RectangleF tmp = new System.Drawing.RectangleF(rect.X, rect.Y, rect.Width, rect.Height);
            return Rectangle.Round(tmp);
        }

        #region Form Events

        private void QuadTreeExample_Resize(object sender, System.EventArgs e)
        {
            Init();
            this.pictureBox.Invalidate();
        }

        #endregion

        #region Picturebox Events

        private void pictureBox_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                _dragging = true;
                _startPoint = e.Location;
            }
        }

        private void pictureBox_MouseMove(object sender, MouseEventArgs e)
        {
            if (_dragging)
            {
                _selectionRect = new QuadTreeLib.RectangleF(
                    Math.Min(e.Location.X, _startPoint.X),
                    Math.Min(e.Location.Y, _startPoint.Y),
                    Math.Abs(e.Location.X - _startPoint.X),
                    Math.Abs(e.Location.Y - _startPoint.Y));

                this.pictureBox.Invalidate();
            }
        }

        private void pictureBox_MouseUp(object sender, MouseEventArgs e)
        {
            if (_dragging && e.Button == MouseButtons.Right)
            {
                foreach (var item in _items)
                    item.Color = Color.Red;

                foreach (var item in _quadTree.Find(_selectionRect))
                    item.Color = Color.GhostWhite;

                foreach (var item in _quadTree.FindIntersecting(_selectionRect))
                    item.Color = Color.DarkMagenta;

                _dragging = false;
            }
            else
            {
                var item = new Item(e.Location.X, e.Location.Y, _index++, 5, 5);
                _items.Add(item);
                _quadTree.Insert(item);
            }

            this.pictureBox.Invalidate();
        }

        private void pictureBox_Paint(object sender, PaintEventArgs e)
        {
            e.Graphics.Clear(Color.Gray);

            _quadTree.ForEach(node => 
            {
                e.Graphics.DrawRectangle(Pens.Black, RoundReactangle(node.Bounds));
            });

            if (!_selectionRect.IsEmpty)
            {
                // draw the selection rect in semi-transparent yellow
                using (Brush b = new SolidBrush(Color.FromArgb(128, Color.Yellow)))
                    e.Graphics.FillRectangle(b, RoundReactangle(_selectionRect));
            }

            foreach (var item in _items)
            {
                e.Graphics.DrawRectangle(new Pen(item.Color), RoundReactangle(item.Rectangle));
            }
        }

        #endregion

        #region Menu Events

        private void AddItems(int count)
        {
            Random rnd = new Random();
            var list = new List<Item>();
            DateTime start = DateTime.Now;

            Debug.WriteLine("Creating {0} items in list", count);

            for (int i = 0; i < count; i++)
            {
                list.Add(new Item(rnd.Next(this.pictureBox.ClientRectangle.Left, this.pictureBox.ClientRectangle.Right - 5),
                    rnd.Next(this.pictureBox.ClientRectangle.Top, this.pictureBox.ClientRectangle.Bottom - 5), _index++, 5, 5));
            }

            Debug.WriteLine("Took {0} ms", (DateTime.Now - start).TotalMilliseconds);
            start = DateTime.Now;
            Debug.WriteLine("Inserting {0} items in Quad tree", count);

            foreach (var item in list)
                _quadTree.Insert(item);

            Debug.WriteLine("Took {0} ms", (DateTime.Now - start).TotalMilliseconds);

            _items.AddRange(list);
        }

        private void add100itemsToolStripMenuItem_Click(object sender, System.EventArgs e)
        {
            AddItems(100);
            this.pictureBox.Invalidate();
        }


        private void add1000ItemsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AddItems(1000);
            this.pictureBox.Invalidate();
        }

        private void clearItemsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Init();
            this.pictureBox.Invalidate();
        }

        private void collisionTestToolStripMenuItem_Click(object sender, EventArgs e)
        {
            var start = DateTime.Now;
            Debug.WriteLine("Performing (n*n)/2 comparison in list");

            int collisions = 0;

            for (int i = 0; i < _items.Count - 1; i++)
            {
                for (int j = i + 1; j < _items.Count; j++)
                {
                    if (_items[i].Rectangle.IntersectsWith(_items[j].Rectangle))
                    {
                        _items[i].Color = Color.Blue;
                        _items[j].Color = Color.Blue;
                        collisions++;
                    }
                }
            }
            Debug.WriteLine("Took {0} ms ({1} collision)", (DateTime.Now - start).TotalMilliseconds, collisions);

            start = DateTime.Now;
            collisions = 0;
            Debug.WriteLine("Performing comparison in Quad tree");

            foreach (var item in _items)
            {
                foreach (var subItem in _quadTree.FindIntersecting(item.Rectangle))
                {
                    if (subItem.Id > item.Id)
                    {
                        item.Color = Color.Green;
                        subItem.Color = Color.Green;
                        collisions++;
                    }
                }

            }
            Debug.WriteLine("Took {0} ms ({1} collision)", (DateTime.Now - start).TotalMilliseconds, collisions);


            this.pictureBox.Invalidate();
        }

        #endregion

    }
}
