﻿using System;
using System.Collections.Generic;

namespace QuadTreeLib
{
    public class QuadTreeNode<T> where T : IHasRectangleF
    {
        public RectangleF Bounds { get; private set; }
        public List<T> Contents { get; private set; }

        protected List<QuadTreeNode<T>> Nodes = new List<QuadTreeNode<T>>(4);
        protected int LevelsLeft;
        protected int SplitSize;

        private bool _hasSplit = false;

        public QuadTreeNode(RectangleF bounds, int levelsLeft, int splitSize)
        {
            _hasSplit = levelsLeft <= 0;
            LevelsLeft = levelsLeft;
            SplitSize = splitSize;
            Contents = new List<T>();
            Bounds = bounds;
        }

        public bool IsEmpty { get { return Bounds.IsEmpty || (Nodes.Count == 0 && Contents.Count == 0); } }

        public int Count
        {
            get
            {
                int count = 0;

                foreach (QuadTreeNode<T> node in Nodes)
                    count += node.Count;

                count += this.Contents.Count;
                return count;
            }
        }

        internal void AddSubTreeContents(List<T> results)
        {
            foreach (QuadTreeNode<T> node in Nodes)
                node.AddSubTreeContents(results);

            results.AddRange(this.Contents);
        }

        internal List<T> Find(RectangleF queryArea, List<T> results)
        {
            results.AddRange(Contents);

            foreach (QuadTreeNode<T> node in Nodes)
            {
                if (node.IsEmpty)
                    continue;

                if (node.Bounds.Contains(queryArea))
                {
                    node.Find(queryArea, results);
                    break;
                }
                if (queryArea.Contains(node.Bounds))
                {
                    node.AddSubTreeContents(results);
                    continue;
                }
                if (node.Bounds.IntersectsWith(queryArea))
                {
                    node.Find(queryArea, results);
                }
            }

            return results;
        }

        internal List<T> FindIntersecting(RectangleF queryArea, List<T> results)
        {
            foreach (T item in this.Contents)
            {
                if (queryArea.IntersectsWith(item.Rectangle))
                    results.Add(item);
            }

            foreach (QuadTreeNode<T> node in Nodes)
            {
                if (node.IsEmpty)
                    continue;

                // Fall 1
                if (node.Bounds.Contains(queryArea))
                {
                    node.FindIntersecting(queryArea, results);
                    break;
                }

                // Fall 2
                if (queryArea.Contains(node.Bounds))
                {
                    node.AddSubTreeContents(results);
                    continue;
                }

                // Fall 3
                if (node.Bounds.IntersectsWith(queryArea))
                {
                    node.FindIntersecting(queryArea, results);
                }
            }

            return results;
        }

        public void Insert(T item)
        {
            if (!Bounds.Contains(item.Rectangle))
                return;

            if (!_hasSplit)
            {
                if (Contents.Count < SplitSize)
                {
                    Contents.Add(item);
                    return;
                }
                else
                {
                    // Kommentar 1
                    CreateSubNodes();

                    var tmpArray = Contents.ToArray();
                    Contents.Clear();

                    foreach (var i in tmpArray)
                        this.Insert(i);
                }
            }
            // Kommerntar 2
            foreach (QuadTreeNode<T> node in Nodes)
            {
                if (node.Bounds.Contains(item.Rectangle))
                {
                    node.Insert(item);
                    return;
                }
            }

            // Kommentar 3
            this.Contents.Add(item);
        }

        public void ForEach(Action<QuadTreeNode<T>> action)
        {
            action(this);

            // draw the child quads
            foreach (var node in this.Nodes)
                node.ForEach(action);
        }

        private void CreateSubNodes()
        {
            _hasSplit = true;

            if (LevelsLeft <= 0)
                return;

            float halfWidth = (Bounds.Width / 2f);
            float halfHeight = (Bounds.Height / 2f);

            Nodes.Add(new QuadTreeNode<T>(new RectangleF(Bounds.X, Bounds.Y, halfWidth, halfHeight), LevelsLeft - 1, SplitSize));
            Nodes.Add(new QuadTreeNode<T>(new RectangleF(Bounds.X, Bounds.Y + halfHeight, halfWidth, halfHeight), LevelsLeft - 1, SplitSize));
            Nodes.Add(new QuadTreeNode<T>(new RectangleF(Bounds.X + halfWidth, Bounds.Y, halfWidth, halfHeight), LevelsLeft - 1, SplitSize));
            Nodes.Add(new QuadTreeNode<T>(new RectangleF(Bounds.X + halfWidth, Bounds.Y + halfHeight, halfWidth, halfHeight), LevelsLeft - 1, SplitSize));
        }
    }
}
